function [MarkovChain,params,options_,results]=AdaMCMC(data,params,options_)

% % % % % % Input:
options_=CreateOptions(options_);
% data: data
% params: name, initial value, min, max, mean, std, target, params.local
% options_




% % % % % % % Output:
%  RESULTS   structure that contains results and information about
%            the simulations
% MarkovChain
%           



modelfun=options_.mh_modelfun;

par0 = params.value(params.parind);


if isempty(options_.mh_incov)
    options_.mh_incov = params.paramsig.^2;
    ii = isinf(options_.mh_incov)|isnan(options_.mh_incov);
    options_.mh_incov(ii) = [abs(params.value(ii))*0.05].^2; % default is 5% std
    options_.mh_incov(options_.mh_incov==0) = 1; % .. or one if we start from zero
    options_.mh_incov = diag(options_.mh_incov);
end

if isempty(options_.mh_adascale)|options_.mh_adascale<=0
    incov_scale = 2.4 / sqrt(options_.mh_npars) ; % scale factor in R
else
    incov_scale = options_.mh_adascale;
end
burnin_scale = 10; % scale in burn-in down/up
incov_adjust  = 1e-5; % eps adjustment


incov = options_.mh_incov(params.parind,params.parind);
R    = chol(options_.mh_incov); % *incov_scale;




oldpar=par0(:)';
params.value(params.parind) = oldpar;
if all(params.local==0)
    ss = feval(options_.mh_ssfun,params.value(:)',data);
else
    ss = feval(options_.mh_ssfun,params.value(:)',data,params.local);
end
ss1 = ss;
ss2 = ss;

ny = length(ss);

if length(options_.mh_S20)==1
    options_.mh_S20 = ones(1,ny)*options_.mh_S20;
end
if length(options_.mh_N)==1
    options_.mh_N = ones(1,ny)*options_.mh_N;
end

if length(options_.mh_N0)==1
    options_.mh_N0 = ones(1,ny)*options_.mh_N0;
end

% default prior function calculates Gaussian sum of squares


oldprior = feval(options_.mh_priorfun,oldpar,params.parammu(params.parind),params.paramsig(params.parind));



% save Chains
MarkovChain   = zeros(options_.mh_replic,options_.mh_npars);
if options_.mh_updatesigma
    s2chain = zeros(options_.mh_replic,ny);
else
    s2chain = [];
end
SS_MarkovChain = zeros(options_.mh_replic,ny);

%% save MarkovChain



MarkovChain(1,:)   = oldpar;
if options_.mh_updatesigma
    s2chain(1,:) = options_.mh_sigma2;
end
SS_MarkovChain(1,:) = ss;

rej=0; reju=0; ii=1; rejl = 0;
%% setup waitbar

% covariance update uses these to store previous values
xcov = []; xmean = []; wsum = []; lasti = 0;

% no update for these indeces
noupd = logical(zeros(1,options_.mh_npars));
noupd(intersect(params.parind,options_.mh_noadaptind)) = 1;

ind = 1; %chain index
si=1;
for i=2:options_.mh_replic % simulation loop

    ii = ii+1; % local adaptation index (?)
    ind = ind+1;

    % waitbar


    % sample new candidate from Gaussian proposal
    newpar=oldpar+randn(1,options_.mh_npars)*R;

    % reject points outside boundaries
    if any(newpar<params.low(params.parind)) | any(newpar>params.upp(params.parind))
        accept = 0;
        MarkovChain(ind,:) = oldpar;
        rej        = rej + 1;
        reju       = reju + 1;

        newprior = 0;
        tst      = 0;
        ss1      = Inf;
        ss2      = ss;
        outbound = 1;
    else
        outbound = 0;
        % prior SS for the new theta
        newprior = feval(options_.mh_priorfun,newpar,params.parammu(params.parind),params.paramsig(params.parind));

        % calculate ss
        ss2 = ss;             % old ss
        params.value(params.parind) = newpar;
        if all(params.local==0)
            ss1 = feval(options_.mh_ssfun,params.value(:)',data);
        else
            ss1 = feval(options_.mh_ssfun,params.value(:)',data,params.local);
        end
        tst = exp(-0.5*( sum((ss1-ss2)./options_.mh_sigma2) + newprior-oldprior) );
        %        p = min( 1, exp([1/kTi - 1/kTj][E i - E j])).
        if tst <= 0
            accept = 0;
            MarkovChain(ind,:) = oldpar;
            rej        = rej + 1;
            reju       = reju + 1;
            if outbound
                rejl     = rejl + 1;
            end
        elseif tst >= 1
            accept = 1;
            MarkovChain(ind,:) = newpar;
            oldpar     = newpar;
            oldprior   = newprior;
            ss         = ss1;

        elseif tst > rand
            accept = 1;
            MarkovChain(ind,:) = newpar;
            oldpar     = newpar;
            oldprior   = newprior;
            ss         = ss1;
        else
            accept = 0;
            MarkovChain(ind,:) = oldpar;
            rej        = rej + 1;
            reju       = reju + 1;
            if outbound
                rejl     = rejl + 1;
            end
        end

    end

    %%% update sigma2


    if options_.mh_updatesigma

        for j=1:ny
            options_.mh_sigma2(j) = 1./((2./(options_.mh_N0(j).*options_.mh_S20(j)+ss(j)))*gammar((options_.mh_N0(j)+options_.mh_N(j))/2));
        end


        s2chain(ind,:) = options_.mh_sigma2;
    end
    %%% Possibly update the prior parameters (for testing hiearchical priors)
    %%% [mu,sig]=priorupdatefun(theta, mu, sig, data)
    if not(isempty(options_.mh_priorupdatefun))
        [params.parammu(params.parind),params.paramsig(params.parind)] = ...
            feval(options_.mh_priorupdatefun,oldpar,params.parammu(params.parind),params.paramsig(params.parind),data);
    end
    %%%
    SS_MarkovChain(ind,:) = ss;
    %
    %% adaptation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if options_.mh_adapint>0 & i<=options_.mh_replic & fix(i/options_.mh_adapint) == i/options_.mh_adapint
        if i < options_.mh_burnintime
            % During burnin no adaptation, just scaling up/down
            if reju/ii > 0.95
                R = R./burnin_scale;
            elseif reju/ii < 0.05
                R = R.*burnin_scale;
            end
        else

            n=options_.mh_adapint;
            p=options_.mh_npars;
            w=ones(n,1);
            wsum0=si*n;
            if si==1
                xmean = zeros(1,p);
                xcov  = zeros(p,p);
            end
            x=MarkovChain((lasti+1):ind,1:options_.mh_npars);
            for ii=1:p
                xmean(ii) = sum(x(:,ii).*w)./wsum0;
            end
            if wsum0>1
                for ii=1:p
                    for jj=1:ii
                        xcov(ii,jj) = (x(:,ii)-xmean(ii))' * ((x(:,jj)-xmean(jj)).*w)./(wsum0-1);
                        if (ii ~= jj)
                            xcov(jj,ii) = xcov(ii,jj);
                        end
                    end
                end
            end
            wsum=wsum0;

            si=si+1;


            lasti = ind;
            %%%
            upcov          = xcov;
            upcov(noupd,:) = options_.mh_incov(noupd,:);
            upcov(:,noupd) = options_.mh_incov(:,noupd);
            %%%
            [Ra,p] = chol(upcov);
            if p % singular cmat
                % try to blow it
                [Ra,p] = chol(upcov + eye(options_.mh_npars)*incov_adjust);
                if p % stil singular
                    R=R;
                else
                    % scale R
                    R = Ra * incov_scale;
                end
            else
                R = Ra * incov_scale;
            end
            lasti = i;

        end
        reju = 0; ii = 0;
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% save MarkovChain

end % replic

% save the rest


params.value(params.parind) = oldpar; % update the initial value to the final value

%% build the results structure
if nargout>0
    results.rejected   = rej/options_.mh_replic;;
    results.ulrejected = rejl/options_.mh_replic;
    results.R      = R;
    results.incov   = R'*R; % with scale %  ./ incov_scale.^2 ;
    qcovorig(params.parind,params.parind) = options_.mh_incov;
    results.incov2  = qcovorig; % original size
    results.cov    = xcov;
    results.mean   = xmean;
    results.params.names  = params.names(params.parind);
    results.limits = [params.low(params.parind)',params.upp(params.parind)'];
    results.prior  = [params.parammu(params.parind)',params.paramsig(params.parind)'];
    results.theta  = params.value; % last values
    results.params.parind = params.parind;
    results.params.local  = params.local;
    results.N      = options_.mh_N;
    if options_.mh_updatesigma
        results.sigma2 = NaN;
        results.S20    = options_.mh_S20;
        results.N0     = options_.mh_N0;
    else
        results.sigma2 = options_.mh_sigma2;
        results.S20    = NaN;
        results.N0     = NaN;
    end
    results.modelfun = options_.mh_modelfun;
    results.ssfun    = options_.mh_ssfun;
    results.priorfun = options_.mh_priorfun;
    results.replic    = options_.mh_replic;
    results.adapint = options_.mh_adapint;
    results.adascale = options_.mh_adascale;

end

